/*
 * Copyright (c) 2008
 * Evan Teran
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appears in all copies and that both the
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the same name not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission. We make no representations about the
 * suitability this software for any purpose. It is provided "as is"
 * without express or implied warranty.
 */

#include <assert.h>

#include <OSGBaseTypes.h>
#include <OSGUInt128.h>

using namespace std;

OSG_BEGIN_NAMESPACE

// constructors for all basic types
UInt128::UInt128(void) :
    lo(0),
    hi(0)
{
}

UInt128::UInt128(Int32 value) :
    lo(static_cast<base_type>(value)),
    hi(0)
{
    if(value < 0)
        hi = static_cast<base_type>(-1);
}

UInt128::UInt128(UInt32 value) :
    lo(static_cast<base_type>(value)),
    hi(0)
{
}

UInt128::UInt128(const UInt128 &value) :
    lo(value.lo),
    hi (value.hi)
{
}

UInt128 &UInt128::operator=(const UInt128 &other)
{
    if(&other != this)
    {
        lo = other.lo; 
        hi = other.hi; 
    }
    return *this;
}

// comparison operators

bool UInt128::operator==(const UInt128 &o) const
{
    return hi == o.hi && lo == o.lo;
}

/*
bool UInt128::operator==(const bool &o) const
{
    return ((hi != 0) || (lo != 0));
}
*/

bool UInt128::operator!=(const UInt128 &o) const
{
    return hi != o.hi || lo != o.lo;
}

/*
bool UInt128::operator!=(const Int64 &o) const
{
    return lo != o;
}

bool UInt128::operator<(const UInt128 &o) const
{
    return (hi == o.hi) ? lo < o.lo : hi < o.hi;
}

bool UInt128::operator>=(const UInt128 &o) const
{
    return !operator<(o);
}

// unary operators

bool UInt128::operator!() const
{
    return !(hi != 0 || lo != 0);
}
*/

/*
UInt128::operator bool() const
{
    return operator!();
}
*/

UInt128 UInt128::operator~() const
{
    UInt128 t(*this);
    t.lo = ~t.lo;
    t.hi = ~t.hi;
    return t;
}

// basic math operators

UInt128 UInt128::operator>>(const Int32 &shift) const
{
    UInt128 result(*this);
    result >>= shift;
    return result;
}

UInt128 UInt128::operator<<(const Int32 &shift) const
{
    UInt128 result(*this);
    result <<= shift;
    return result;
}

UInt128 UInt128::operator&(const UInt128 &b)
{
    UInt128 result(*this);
    result &= b;
    return result;
}

UInt128 UInt128::operator|(const UInt128 &b)
{
    UInt128 result(*this);
    result |= b;
    return result;
}

UInt128 UInt128::operator^(const UInt128 &b)
{
    UInt128 result(*this);
    result ^= b;
    return result;
}

UInt128 &UInt128::operator|=(const UInt128 &b)
{
    hi |= b.hi;
    lo |= b.lo;
    return *this;
}

UInt128 &UInt128::operator&=(const UInt128 &b)
{
    hi &= b.hi;
    lo &= b.lo;
    return *this;
}

UInt128 &UInt128::operator^=(const UInt128 &b)
{
    hi ^= b.hi;
    lo ^= b.lo;
    return *this;
}

UInt128 &UInt128::operator<<=(const Int32 &shift)
{
    Int32 n = shift;
    if(n >= size)
    {
        hi = 0;
        lo = 0;
    }
    else
    {
        const unsigned int halfsize = size / 2;

        if (n >= halfsize)
        {
            n -= halfsize;
            hi = lo;
            lo = 0;
        }

        if (n != 0)
        {
            // shift high half
            hi <<= n;

            const base_type mask(~(base_type(-1) >> n));

            // and add them to high half
            hi |= (lo & mask) >> (halfsize - n);

            // and finally shift also low half
            lo <<= n;
        }
    }
    
    return *this;
}

UInt128 &UInt128::operator>>=(const Int32 &shift)
{
    Int32 n = shift;
    if(n >= size)
    {
        hi = 0;
        lo = 0;
    }
    else
    {
        const unsigned int halfsize = size / 2;

        if (n >= halfsize)
        {
            n -= halfsize;
            lo = hi;
            hi = 0;
        }

        if (n != 0)
        {
            // shift low half
            lo >>= n;

            // get lower N bits of high half
            const base_type mask(~(base_type(-1) << n));

            // and add them to low qword
            lo |= (hi & mask) << (halfsize - n);

            // and finally shift also high half
            hi >>= n;
        }
    }

    return *this;
}

const UInt128 TypeTraits<UInt128>::Zero = 0;
const UInt128 TypeTraits<UInt128>::One = 1;


OSG_END_NAMESPACE

/*-------------------------------------------------------------------------*/
/*                              cvs id's                                   */

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static OSG::Char8 cvsid_cpp[] = "@(#)$Id: $";
    static OSG::Char8 cvsid_hpp[] = OSGUINT128_HEADER_CVSID;
}
