/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/


#ifndef _GROUPCONNECTION_H_
#define _GROUPCONNECTION_H_
#ifdef __sgi
#pragma once
#endif

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------

#include <string>
#include <set>
#include <OSGBase.h>
#include <OSGBaseTypes.h>
#include <OSGConnection.h>
#include <OSGTime.h>

OSG_BEGIN_NAMESPACE

class OSG_BASE_DLLMAPPING GroupConnection : public Connection
{
    /*==========================  PUBLIC  =================================*/
  public:

    /*---------------------------------------------------------------------*/
    /*! \name                   Constructors                               */
    /*! \{                                                                 */

             GroupConnection ( int zeroCopyThreshold );
    virtual ~GroupConnection ( void                  );

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   type info                                  */
    /*! \{                                                                 */

    virtual const ConnectionType *getType (void) const = 0;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   connection                                 */
    /*! \{                                                                 */

    virtual void    disconnect  (      Channel     channel   ) = 0;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   channel handling                           */
    /*! \{                                                                 */

    UInt32  getChannelCount  (void           );
    void    addSelection     (Channel channel);
    void    subSelection     (Channel channel);
    void    clearSelection   (void           );
    void    resetSelection   (void           );
    UInt32  getSelectionCount(void           );

    virtual Channel selectSingleChannel(Channel channel, Time timeout=-1) throw (ReadError) = 0;
    virtual Int32 selectSingleChannelAndGetAvaiable(Channel channel, Time timeout) throw (ReadError) = 0;
    virtual bool hasAvaiable(Channel channel, Int32 timeout) throw (ReadError) = 0;
    virtual int getAvailable(Channel channel) = 0;
    virtual int waitAvailable(Channel channel) = 0;

    virtual std::vector<Channel> selectChannels(Time timeout=-1) throw (ReadError) = 0;

    virtual SizeT read(MemoryHandle mem, SizeT size) = 0;
    virtual SizeT read(Channel channel, MemoryHandle mem, SizeT size) = 0;

    virtual void getFromChannel        (Channel channel, void         *dst,   UInt32 size) = 0;
    virtual void getValueFromChannel   (Channel channel, bool         &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, UInt8        &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, UInt16       &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, UInt32       &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, UInt64       &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, Int8         &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, Int16        &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, Int32        &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, Int64        &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, Real16       &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, Real32       &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, Real64       &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, Real128      &value             ) = 0;
    virtual void getValueFromChannel   (Channel channel, std::string  &value             ) = 0;

    virtual void write                 (const MemoryHandle mem, SizeT size) = 0;
    virtual void write                 (Channel channel, const MemoryHandle mem, SizeT size) = 0;

    virtual void putToChannel       (Channel channel, void const   *src,   UInt32 size ) = 0;
    virtual void putValueToChannel  (Channel channel, const bool         &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const UInt8        &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const UInt16       &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const UInt32       &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const UInt64       &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const Int8         &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const Int16        &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const Int32        &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const Int64        &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const Real16       &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const Real32       &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const Real64       &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const Real128      &value        ) = 0;
    virtual void putValueToChannel  (Channel channel, const std::string  &value        ) = 0;

    virtual void resetWriteBufferStatistic(void);
    virtual void getWriteBufferStatistic(Real64 &t, UInt64 &bytes, UInt64 &cbytes);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   group address                              */
    /*! \{                                                                 */

    void        setDestination(const std::string &destination);
    std::string getDestination(void                          );

    /*! \}                                                                 */

    /*=========================  PROTECTED  ===============================*/
  protected:
    /*---------------------------------------------------------------------*/
    /*! \name                   protected fields                           */
    /*! \{                                                                 */

    std::vector<UInt8>        _selection;
    std::string               _destination;
    std::set<Channel>         _disconnectedChannel;

    Real64                    _write_buffer_time;
    UInt64                    _write_buffer_bytes;
    UInt64                    _write_buffer_cbytes;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   protected types                            */
    /*! \{                                                                 */

    typedef Int32 ChannelIndex;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   internal channel handling                  */
    /*! \{                                                                 */

    Channel newChannelIndex(ChannelIndex index);
    void    delChannelIndex(ChannelIndex index);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   channel index mapping                      */
    /*! \{                                                                 */

    ChannelIndex channelToIndex (Channel      channel ) const;
    Channel      indexToChannel (ChannelIndex index   ) const;

    /*! \}                                                                 */

    /*==========================  PRIVATE  ================================*/
  private:
    /*---------------------------------------------------------------------*/
    /*! \name                   private fields                             */
    /*! \{                                                                 */

    std::vector<ChannelIndex> _channelToIndex;
    std::vector<Channel>      _indexToChannel;
    std::list<Channel>        _reuseChannel;

    /*! \}                                                                 */
    typedef Connection Inherited;

    // prohibit default functions (move to 'public' if you need one)
    GroupConnection(const GroupConnection &source);
    GroupConnection& operator =(const GroupConnection &source);
};

//---------------------------------------------------------------------------
//   Exported Types
//---------------------------------------------------------------------------

// class pointer

typedef GroupConnection *GroupConnectionP;

OSG_END_NAMESPACE

#define OSG_GROUPCONNECTION_HEADER_CVSID "@(#)$Id: $"

#endif /* _CONNECTION_H_ */
