/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#ifndef _DGRAM_H_
#define _DGRAM_H_
#ifdef __sgi
#pragma once
#endif

#include <list>
#include <vector>

#include <OSGBaseTypes.h>
#include <OSGTime.h>
#include <OSGBase.h>
#include <OSGLock.h>
#include <OSGBaseFunctions.h>

OSG_BEGIN_NAMESPACE

class OSG_BASE_DLLMAPPING Dgram
{
    /*==========================  PUBLIC  =================================*/
public:

    enum { DGRAM_UNKNOWN = 0,
           DGRAM_COMPRESSED = 1,
           DGRAM_LAST = 2 };
    /*---------------------------------------------------------------------*/
    /*! \name  Constants                                                   */
    /*! \{                                                                 */

    // max size is 65507 bytes!
    //static const UInt32 DgramLength = 9216;
    //static const UInt32 DgramLength = 1472;
    // I send each packet twice so 1472 / 2 seems to be the best size.
    static const UInt32 DgramLength = 736;
    //static const UInt32 DgramLength = 16000;

    /*---------------------------------------------------------------------*/
    /*! \name               constructor                            */
    /*! \{                                                                 */

    Dgram();
    ~Dgram();

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name               get                                            */
    /*! \{                                                                 */

    inline       UInt32         getCapacity      (void) const;
    inline       UInt16         getId            (void) const;
    inline       UInt16         getBaseId        (void) const;
    inline       UInt16         getFlags         (void) const;
    inline       bool           getResponseAck   (void) const;
    inline       UInt32         getSize          (void) const;
    inline       UChar8        *getData          (void);
    inline const UChar8        *getData          (void) const;

    inline       UChar8        *getBuffer        (void);
    inline       UInt32         getBufferSize    (void) const;
    inline       UInt32         getBufferCapacity(void) const;

    inline       bool           getEarlySend     (void) const;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name               set                                            */
    /*! \{                                                                 */

    inline void setSize         (UInt32 size    );
    inline void setResponseSize (void           );
    inline void setBaseId       (UInt16 bid     );
    inline void setId           (UInt16 id      );
    inline void setFlags        (UInt16 flags   );
    inline void setResponseAck  (bool value     );

    inline void setBufferSize   (UInt32 size    );

    inline void setEarlySend    (bool value     );

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name               compare                                        */
    /*! \{                                                                 */

    inline        bool operator < (const Dgram  &other) const;
    inline static bool less(UInt16,UInt16);

    /*! \}                                                                 */
    /*=========================  PROTECTED  ===============================*/
protected:

    /*---------------------------------------------------------------------*/
    /*! \name               member                                         */
    /*! \{                                                                 */

    /*! \nohierarchy
     */
    struct Buffer
    {
        UInt16 _bid;
        UInt16 _id;
        UInt16 _flags;
        UChar8 _data[DgramLength - 6];
    };

    bool   _earlySend;
    UInt32 _size;
    Buffer _buffer;

    /*! \}                                                                 */
    /*==========================  PRIVATE  ================================*/
private:

};

// class pointer
typedef Dgram *DgramP;

OSG_END_NAMESPACE

#include "OSGDgram.inl"

#define OSG_DGRAMHEADER_CVSID "@(#)$Id:$"

#endif /* _DGRAM_H_ */
