/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

// amz removed index to channel mapping
// leads to wrong mappings after disconnecting a client!
// TODO remove the whole channel concept? Just use index?

//#define MULTI_THREADED


//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include <sstream>

#include "OSGConfig.h"
#include "OSGLog.h"
#include "OSGBaseFunctions.h"
#include "OSGSocketSelection.h"
#include "OSGSocketException.h"
#include "OSGGroupSockConnection.h"
#include "OSGConnectionHelper.h"
#include "OSGConnectionType.h"

#include "OSGminiz.h"
#include "OSGLZ4.h"

#pragma intrinsic(memcpy)

#ifdef WIN32
#define POLLRDNORM 0x0100
typedef struct pollfd
{
    SOCKET fd;
    short  events;
    short  revents;
} WSAPOLLFD, *PWSAPOLLFD, *LPWSAPOLLFD;
typedef int (/*WSAAPI*/ * WSAPollFT)(WSAPOLLFD fdarray[], ULONG nfds, INT timeout);
static WSAPollFT WSAPoll = NULL;
#else
#include <poll.h>
#endif

OSG_USING_NAMESPACE

/** \class osg::GroupSockConnection
 *  \ingroup GrpBaseNetwork
 **/

// amz changed from 131071 to 131072
/* static */ const UInt32 GroupSockConnection::_ioBufferSize = 131072;

/*-------------------------------------------------------------------------*/
/*                            constructor destructor                       */

/*! Constructor
 */

GroupSockConnection::GroupSockConnection():
    GroupConnection(0),
    _acceptSocket(),
    _sockets(),
    _remoteAddresses(),
    _readIndex(0),
    _socketReadBuffer(),
    _socketWriteBuffer(),
    _tp(),
    _id(0)
{
    _acceptSocket.open();
    _acceptSocket.setReusePort(true);

    _socketReadBuffer .resize(_ioBufferSize);
    _socketWriteBuffer.resize(_ioBufferSize);
    // reserve first bytes for buffer size
    readBufAdd (&_socketReadBuffer [sizeof(SocketBufferHeader)],
                _socketReadBuffer.size() - sizeof(SocketBufferHeader));
    writeBufAdd(&_socketWriteBuffer[sizeof(SocketBufferHeader)],
                _socketWriteBuffer.size() - sizeof(SocketBufferHeader));

#ifdef WIN32
    HINSTANCE ws2 = ::GetModuleHandle("Ws2_32.dll");
    WSAPoll =  (WSAPollFT) (void *) GetProcAddress(ws2, "WSAPoll");
#endif
}

/*! Destructor
 */
GroupSockConnection::~GroupSockConnection(void)
{
    // close and remove sockets
    while(_sockets.size())
    {
        try
        {
            _sockets.back    ().close();
            _sockets.pop_back();
        }
        catch(...)
        {
        }
    }
    _acceptSocket.close();
}

/*! get connection type
 */
const ConnectionType *GroupSockConnection::getType(void) const
{
    return &_type;
}

/*-------------------------------------------------------------------------*/
/*                            connection                                   */

/*! Connect to the given point. If timeout is reached, -1 is
    returned
 */
GroupConnection::Channel GroupSockConnection::connectPoint(
    const std::string &address,
    Time               timeout)
{
    Channel channel = -1;
    StreamSocket socket;
    SocketAddress destination;
    if(connectSocket(socket, address, destination, timeout))
    {
        channel = newChannelIndex(ChannelIndex(_sockets.size()));
        _sockets        .push_back(socket);
        _remoteAddresses.push_back(destination);
        _readIndex = 0;
        setLocalInterface(socket.getAddress().getHost());
        setRemoteInterface(destination.getHost());
    }
    else
    {
        if(_sockets.empty())
            setLocalInterface("");
    }
    return channel;
}

/*! Accept an icomming point connection. If timeout is reached,
    -1 is returned. If timeout is -1 then wait without timeout
 */
GroupConnection::Channel GroupSockConnection::acceptPoint(Time timeout)
{
    StreamSocket from;
    SocketAddress destination;
    if(acceptSocket(_acceptSocket, from,
                    destination,   timeout))
    {
        Channel channel = newChannelIndex(ChannelIndex(_sockets.size()));
        _sockets        .push_back(from);
        _remoteAddresses.push_back(destination);
        _readIndex = 0;
        setLocalInterface(from.getAddress().getHost());
        setRemoteInterface(destination.getHost());
        return channel;
    }
    else
    {
        if(_sockets.empty())
        {
            setLocalInterface("");
            setRemoteInterface("");
        }
        return -1;
    }
}

/*! Disconnect the given channel
 */
void GroupSockConnection::disconnect(Channel channel)
{
    ChannelIndex index = channelToIndex(channel);
    //printf("disconnect channel %d index %d\n", channel, index);
    try
    {
        _sockets[index].close();
    }
    catch(...)
    {
    }

    _sockets        .erase(_sockets        .begin() + index);
    _remoteAddresses.erase(_remoteAddresses.begin() + index);
    delChannelIndex(index);
    _readIndex = 0;
}

/*! Bind the connection to an network interface. The address is
    returned, on wich the port could be connected. The interface
    is determined by the connection interface filed and the
    address parameter. Address can be empty, wich means to use
    a free port or address can contain a port number.
 */
std::string GroupSockConnection::bind(const std::string &address)
{
    int         port = 0;
    char        localhost[256];
    char        host[256];
    char        portStr[256];
    std::string interf;
    std::string boundedAddress;

    // get local host name
    osgGetHostname(localhost, 255);

    if(!getInterface().empty())
        interf = getInterface();
    else
        interf = localhost;
    // parse address
    if(!address.empty())
        if(sscanf(address.c_str(), "%*[^:]:%d", &port) != 1)
            if(sscanf(address.c_str(), ":%d", &port) != 1)
                port = 0;
    // bind port
    _acceptSocket.setReusePort(true);
    _acceptSocket.bind(SocketAddress(interf.c_str(), port));
    SINFO << "Connection bound to "
          << _acceptSocket.getAddress().getHost() << ":"
          << _acceptSocket.getAddress().getPort() << std::endl;
    _acceptSocket.listen();
    // create address
    sprintf(portStr, "%d", _acceptSocket.getAddress().getPort());
    return interf + ":" + portStr;
}

/*! Parse the params string.
 */
void GroupSockConnection::setParams(const std::string &params)
{
    if(params.empty())
        return;

    std::string option = "bufferSize=";
    std::string::size_type i = 0;
    if((i = params.find(option)) != std::string::npos)
    {
        std::string str = params.substr(i + option.size());

        std::stringstream ss;
        std::string::size_type j = 0;
        while(j < str.length() && str[j] != ',' && isdigit(str[j]))
        {
            ss << str[j++];
        }
        UInt32 bufferSize;
        ss >> bufferSize;

        // clear old buffer.
        readBufClear();
        writeBufClear();

        _socketReadBuffer.resize(bufferSize);
        _socketWriteBuffer.resize(_socketReadBuffer.size());

        // reserve first bytes for buffer size
        readBufAdd (&_socketReadBuffer [sizeof(SocketBufferHeader)],
                    _socketReadBuffer.size() - sizeof(SocketBufferHeader));
        writeBufAdd(&_socketWriteBuffer[sizeof(SocketBufferHeader)],
                    _socketWriteBuffer.size() - sizeof(SocketBufferHeader));

        FINFO(("GroupSockConnection::setParams : setting buffer size to %u.\n", bufferSize));
    }
}

/*-------------------------------------------------------------------------*/
/*                              channel handling                           */

/*! select the next channel for reading. If timeout is not -1
    then -1 is returned if timeout is reached
*/
Connection::Channel GroupSockConnection::selectChannel(Time timeout)
throw (ReadError)
{
    Int32 maxnread = 0, nread;
    ChannelIndex index;
    SocketSelection selection, result;

    // if there is data in the read buffer, return current channel
    if(_zeroCopyThreshold != 1 &&
            _currentReadBuffer != readBufEnd())
    {
        FFATAL(("Channel change ignores data in current buffer"))
        return indexToChannel(_readIndex);
    }

    if(_selection[_readIndex] &&
            _sockets[_readIndex].getAvailable())
    {
        return indexToChannel(_readIndex);;
    }

    // wait for first socket to deliver data
    for(index = 0 ; index < _sockets.size() ; ++index)
    {
        if(_selection[index])
            selection.setRead(_sockets[index]);
    }

    try
    {
        // select ok ?
        if(!selection.select(timeout, result))
            return -1;

        // use socket with most data
        for(index = 0 ; index < _sockets.size() ; ++index)
        {
            if(result.isSetRead(_sockets[index]))
            {
                nread = _sockets[index].getAvailable();
                if(maxnread < nread)
                {
                    maxnread = nread;
                    _readIndex = index;
                }
            }
        }
    }
    catch(SocketException &e)
    {
        throw ReadError(e.what());
    }

    // return channel id
    return indexToChannel(_readIndex);
}

int GroupSockConnection::getAvailable(void)
{
    return _sockets[_readIndex].getAvailable();
}

int GroupSockConnection::getAvailable(Channel channel)
{
    return _sockets[channelToIndex(channel)].getAvailable();
}

int GroupSockConnection::waitAvailable(Channel channel)
{
    return _sockets[channelToIndex(channel)].waitAvailable();
}

Connection::Channel GroupSockConnection::selectSingleChannel(Channel channel, Time timeout) throw (ReadError)
{
    _readIndex = channelToIndex(channel);

    if(_selection[_readIndex] &&
            _sockets[_readIndex].getAvailable())
    {
        return channel;
    }

    SocketSelection selection;
    if(_selection[_readIndex])
        selection.setRead(_sockets[_readIndex]);

    try
    {
        // select ok ?
        SocketSelection result;
        if(!selection.select(timeout, result))
            return -1;
    }
    catch(SocketException &e)
    {
        throw ReadError(e.what());
    }

    // return channel id
    return channel;
}

Int32 GroupSockConnection::selectSingleChannelAndGetAvaiable(Channel channel, Time timeout) throw (ReadError)
{
    ChannelIndex readIndex = channelToIndex(channel);

    if(_selection[readIndex] &&
            _sockets[readIndex].getAvailable())
    {
        return _sockets[readIndex].getAvailable();
    }

    SocketSelection selection;
    if(_selection[readIndex])
        selection.setRead(_sockets[readIndex]);

    try
    {
        // select ok ?
        SocketSelection result;
        if(!selection.select(timeout, result))
            return -1;
    }
    catch(SocketException &e)
    {
        throw ReadError(e.what());
    }

    // return channel id
    return _sockets[readIndex].getAvailable();
}

// time out in milliseconds.
bool GroupSockConnection::hasAvaiable(Channel channel, Int32 timeout) throw (ReadError)
{
    ChannelIndex readIndex = channelToIndex(channel);

    if(_selection[readIndex] &&
            _sockets[readIndex].getAvailable())
    {
        return true;
    }

    if(timeout == 0)
        return false;

    int sd = _sockets[readIndex].getDescriptor();

#ifdef WIN32

    if(WSAPoll == NULL)
    {
        struct fd_set read;
        FD_ZERO(&read);
        FD_SET(sd, &read);

        timeval tVal, *tValP;
        if(timeout < 0)
        {
            tValP = NULL;
        }
        else
        {
            tVal.tv_sec  = int( timeout / 1000 );
            tVal.tv_usec = int( (timeout - tVal.tv_sec * 1000) * 1000);
            tValP = &tVal;
        }

        // sd + 1 or FD_SETSIZE
        // on windows the first parameter of select is ignored!
        int count = ::select(FD_SETSIZE, &read, NULL, NULL, tValP);
        if(count < 0)
            throw SocketError("hasAvaiable()");

        return (count > 0);
    }
    else
    {
        WSAPOLLFD pfd;
        pfd.fd = sd;
        pfd.events = POLLRDNORM;
        pfd.revents = 0;

        if(WSAPoll(&pfd, 1, timeout) > 0)
            return (pfd.revents & POLLRDNORM);
    }
    return false;

#else

    struct pollfd pfd;
    pfd.fd = sd;
    pfd.events = POLLRDNORM;
    pfd.revents = 0;
    if(::poll(&pfd, 1, timeout) > 0)
        return (pfd.revents & POLLRDNORM);
    return false;

#endif
}

std::vector<Connection::Channel> GroupSockConnection::selectChannels(Time timeout) throw (ReadError)
{
    std::vector<Connection::Channel> channels;

    SocketSelection selection;
    for(UInt32 index = 0; index < _sockets.size() ; ++index)
    {
        if(_selection[index])
            selection.setRead(_sockets[index]);
    }

    try
    {
        SocketSelection result;
        if(!selection.select(timeout, result))
            return channels;

        for(UInt32 index = 0; index < _sockets.size(); ++index)
        {
            if(result.isSetRead(_sockets[index]))
                channels.push_back(indexToChannel(index));
        }
    }

    catch(SocketException &e)
    {
        throw ReadError(e.what());
    }

    // return channel ids
    return channels;
}

void GroupSockConnection::flush(void)
{
    GroupConnection::flush();
#ifdef MULTI_THREADED
    _tp.wait();
#endif
}

void GroupSockConnection::setDelay(bool s)
{
    for(UInt32 index = 0; index < _sockets.size(); ++index)
        _sockets[index].setDelay(s);
}

/*-------------------------------------------------------------------------*/
/*                            sync                                         */

/*! wait for signal
 */
bool GroupSockConnection::wait(Time timeout) throw (ReadError)
{
    UInt32 len;
    UInt32 index;
    UInt32 tag = 314156;
    UInt32 missing = _sockets.size();
    SocketSelection selection, result;

    for(index = 0 ; index < _sockets.size() ; ++index)
        selection.setRead(_sockets[index]);

    try
    {
        std::vector<Int32> remove;
        while(missing)
        {
            if(!selection.select(timeout, result))
                return false;
            for(index = 0 ; index < _sockets.size() ; ++index)
            {
                if(result.isSetRead(_sockets[index]))
                {
                    try
                    {
                        len = _sockets[index].recv(&tag, sizeof(tag));
                    }
                    catch(SocketException &e)
                    {
                        selection.clearRead(_sockets[index]);
                        remove.push_back(indexToChannel(index));
                        missing--;
                        continue;
                    }
                    tag = osgntohl(tag);
                    if(len == 0)
                        throw ReadError("Channel closed");
                    selection.clearRead(_sockets[index]);
                    missing--;
                    if(tag != 314156)
                    {
                        FFATAL(("Stream out of sync in SockConnection (tag is %u)\n", tag));
                        throw ReadError("Stream out of sync");
                    }
                }
            }
        }

        if(!remove.empty())
        {
            for(size_t i=0;i<remove.size();++i)
            {
                disconnect(remove[i]);
            }
            if(getChannelCount() == 0)
            {
                throw ReadError("Last running cluster server crashed!");
            }
        }
    }
    catch(SocketException &e)
    {
        throw ReadError(e.what());
    }
    return true;
}

/*! send signal
 */
void GroupSockConnection::signal(void) throw (WriteError)
{
    UInt32 tag = osghtonl(314156);
    UInt32 index;

    std::vector<Int32> remove;
    for(index = 0 ; index < _sockets.size() ; ++index)
    {
        try
        {
            _sockets[index].send(&tag, sizeof(tag));
        }
        catch(SocketException &e)
        {
            remove.push_back(indexToChannel(index));
        }
    }

    if(!remove.empty())
    {
        for(size_t i=0;i<remove.size();++i)
        {
            disconnect(remove[i]);
        }
        if(getChannelCount() == 0)
        {
            throw WriteError("Last running cluster server crashed!");
        }
    }
}

/*-------------------------- create ---------------------------------------*/

/** \brief create conneciton
 */

GroupConnection *GroupSockConnection::create(void)
{
    return new GroupSockConnection();
}

/*-------------------------------------------------------------------------*/
/*                              read write                                 */

/** Read data into given memory
 *
 * Read data form the current read socket. The read socket is that
 * socket, that was selectet in selectChannel.
 *
 **/

SizeT GroupSockConnection::read(MemoryHandle mem, SizeT size)
{
    int len = 0;

    // read data
    len = _sockets[_readIndex].recv(mem, size);
    if(len == 0)
    {
//        throw ChannelClosed(indexToChannel(_readIndex));
        throw ReadError("Channel closed");
    }

    return len;
}

SizeT GroupSockConnection::read(Channel channel, MemoryHandle mem, SizeT size)
{
    ChannelIndex readIndex = channelToIndex(channel);
    int len = 0;

    // read data
    len = _sockets[readIndex].recv(mem, size);
    if(len == 0)
    {
//        throw ChannelClosed(indexToChannel(_readIndex));
        throw ReadError("Channel closed");
    }

    return len;
}

void GroupSockConnection::getFromChannel(Channel channel, void *dst, UInt32 size)
{
    MemoryHandle data = static_cast<MemoryHandle>(dst);
    read(channel, data, size);
}

void GroupSockConnection::getValueFromChannel(Channel channel, bool &value)
{
    UInt8 temp;
    getFromChannel(channel, &temp, sizeof(UInt8));
    value = (temp != 0);
}

void GroupSockConnection::getValueFromChannel(Channel channel, UInt8 &value)
{
    getFromChannel(channel, &value, sizeof(UInt8));
}

void GroupSockConnection::getValueFromChannel(Channel channel, UInt16 &value)
{
    getFromChannel(channel, &value, sizeof(UInt16));
    value = osgntohs(value);
}

void GroupSockConnection::getValueFromChannel(Channel channel, UInt32 &value)
{
    getFromChannel(channel, &value, sizeof(UInt32));
    value = osgntohl(value);
}

void GroupSockConnection::getValueFromChannel(Channel channel, UInt64 &value)
{
    getFromChannel(channel, &value, sizeof(UInt64));
    value = osgntohll(value);
}

void GroupSockConnection::getValueFromChannel(Channel channel, Int8 &value)
{
    getFromChannel(channel, &value, sizeof(Int8));
}

void GroupSockConnection::getValueFromChannel(Channel channel, Int16 &value)
{
    getFromChannel(channel, &value, sizeof(Int16));
    value = osgntohs(value);
}

void GroupSockConnection::getValueFromChannel(Channel channel, Int32 &value)
{
    getFromChannel(channel, &value, sizeof(Int32));
    value = osgntohl(value);
}

void GroupSockConnection::getValueFromChannel(Channel channel, Int64 &value)
{
    getFromChannel(channel, &value, sizeof(Int64));
    value = osgntohll(value);
}

void GroupSockConnection::getValueFromChannel(Channel channel, Real16 &value)
{
    UInt16 v;
    getFromChannel(channel, &v, sizeof(Real16));
    v     = osgntohs(v);
    value.setBits(v);
}

void GroupSockConnection::getValueFromChannel(Channel channel, Real32 &value)
{
    getFromChannel(channel, &value, sizeof(Real32));
    value = osgntohf(value);
}

void GroupSockConnection::getValueFromChannel(Channel channel, Real64 &value)
{
    getFromChannel(channel, &value, sizeof(Real64));
    value = osgntohd(value);
}

void GroupSockConnection::getValueFromChannel(Channel channel, Real128 &value)
{
    getFromChannel(channel, &value, sizeof(Real128));
    value = osgntohdd(value);
}

void GroupSockConnection::getValueFromChannel(Channel channel, std::string &value)
{
    UInt32  len;
    Char8  *str = NULL;

    getValueFromChannel(channel, len);

    if(len != 0)
    {
        str = new Char8[len];

        getFromChannel(channel, str, len);

        value = str;

        delete [] str;
    }
    else
    {
        value.erase();
    }
}

/** Read next data block
 *
 * The stream connection uses only BinaryDataHandler buffer. If more
 * then one buffer is present, then this methode must be changed!
 *
 */

void GroupSockConnection::readBuffer()
{
    int size;
    int len;

    // read buffer header
    len = _sockets[_readIndex].recv(&_socketReadBuffer[0], sizeof(SocketBufferHeader));
    if(len == 0)
        throw ReadError("Channel closed");

    // read remaining data
    size = osgntohl(reinterpret_cast<SocketBufferHeader *>(&_socketReadBuffer[0])->size);
    UInt32 rsize = size;

    int compressed_size = osgntohl(reinterpret_cast<SocketBufferHeader *>(&_socketReadBuffer[0])->compressed_size);
    if(compressed_size > 0)
        rsize = compressed_size;

    len = _sockets[_readIndex].recv(&_socketReadBuffer[sizeof(SocketBufferHeader)], rsize);
    if(len == 0)
        throw ReadError("Channel closed");

    if(_compression == Connection::CCOMPRESSION_BEST && compressed_size > 0)
    {
        static std::vector<UInt8> mem;
        mem.resize(rsize);
        memcpy(&mem[0], &_socketReadBuffer[sizeof(SocketBufferHeader)], rsize);
        mz_ulong ucsize = _ioBufferSize;
        int r = mz_uncompress((unsigned char *) &_socketReadBuffer[sizeof(SocketBufferHeader)], &ucsize, (const unsigned char *) &mem[0], mem.size());
        if(r != 0)
            printf("GroupsockConnection::readBuffer:: uncompress error!\n");
        size = ucsize;
    }
    else if(_compression == Connection::CCOMPRESSION_FAST && compressed_size > 0)
    {
        static std::vector<UInt8> mem;
        mem.resize(rsize);
        memcpy(&mem[0], &_socketReadBuffer[sizeof(SocketBufferHeader)], rsize);
        size = LZ4_decompress_safe((const char *) &mem[0], (char *) &_socketReadBuffer[sizeof(SocketBufferHeader)], rsize, size);

        //printf("PointSockConnection::readBuffer: uncompressed %d ==> %d\n", len, size);
    }
    else
    {
        //printf("GroupSockConnection::readBuffer: uncompressed %d\n", len);
    }

    readBufBegin()->setDataSize(size);
}

/** Write data to all destinations
 *
 * \param mem   Pointer to data buffer
 * \param size  Size of bytes to write
 *
 **/

void GroupSockConnection::write(const MemoryHandle mem, SizeT size)
{
    if(size == 0)
        return;

    std::vector<Int32> remove;
    // write to all connected sockets
    for(Int32 index = 0 ; index < _sockets.size() ; ++index)
    {
        try
        {
            _sockets[index].send(mem, size);
        }
        catch(SocketException &e)
        {
            remove.push_back(indexToChannel(index));
        }
    }

    if(!remove.empty())
    {
        for(size_t i=0;i<remove.size();++i)
        {
            disconnect(remove[i]);
        }
        if(getChannelCount() == 0)
        {
            throw WriteError("Last running cluster server crashed!");
        }
    }
}

void GroupSockConnection::write(Channel channel, const MemoryHandle mem, SizeT size)
{
    ChannelIndex index = channelToIndex(channel);

    int wbytes = 0;
    try
    {
        wbytes = _sockets[index].send(mem, size);
    }
    catch(SocketException &e)
    {
        throw WriteError(e.what());
    }
}


void GroupSockConnection::putToChannel(Channel channel, void const *src, UInt32 size)
{
    UInt8 const *data = static_cast<UInt8 const *>(src);
    write(channel, const_cast<MemoryHandle>(data), size);
}

void GroupSockConnection::putValueToChannel(Channel channel, const bool &value)
{
    UInt8 temp = UInt8(value);
    putToChannel(channel, &temp, sizeof(UInt8));
}

void GroupSockConnection::putValueToChannel(Channel channel, const UInt8 &value)
{
    putToChannel(channel, &value, sizeof(UInt8));
}

void GroupSockConnection::putValueToChannel(Channel channel, const UInt16 &value)
{
    UInt16 z = osghtons(value);
    putToChannel(channel, &z, sizeof(UInt16));
}

void GroupSockConnection::putValueToChannel(Channel channel, const UInt32 &value)
{
    UInt32 z = osghtonl(value);
    putToChannel(channel, &z, sizeof(UInt32));
}

void GroupSockConnection::putValueToChannel(Channel channel, const UInt64 &value)
{
    UInt64 z = osghtonll(value);
    putToChannel(channel, &z, sizeof(UInt64));
}

void GroupSockConnection::putValueToChannel(Channel channel, const Int8 &value)
{
    putToChannel(channel, &value, sizeof(Int8));
}

void GroupSockConnection::putValueToChannel(Channel channel, const Int16 &value)
{
    Int16 z = osghtons(value);
    putToChannel(channel, &z, sizeof(Int16));
}

void GroupSockConnection::putValueToChannel(Channel channel, const Int32 &value)
{
    Int32 z = osghtonl(value);
    putToChannel(channel, &z, sizeof(Int32));
}

void GroupSockConnection::putValueToChannel(Channel channel, const Int64 &value)
{
    Int64 z = osghtonll(value);
    putToChannel(channel, &z, sizeof(Int64));
}

void GroupSockConnection::putValueToChannel(Channel channel, const Real16 &value)
{
    UInt16 v = osghtons(value.bits());
    putToChannel(channel, &v, sizeof(Real16));
}

void GroupSockConnection::putValueToChannel(Channel channel, const Real32 &value)
{
    UInt32 v = osghtonl( *(reinterpret_cast<const UInt32 *>(&value)) );
    putToChannel(channel, &v, sizeof(Real32));
}

void GroupSockConnection::putValueToChannel(Channel channel, const Real64 &value)
{
    UInt64 v = osghtonll( *(reinterpret_cast<const UInt64 *>(&value)) );
    putToChannel(channel, &v, sizeof(Real64));
}

void GroupSockConnection::putValueToChannel(Channel channel, const Real128 &value)
{
    UInt64 v = osghtonll( *( reinterpret_cast<const UInt64 *>(&value)) );
    UInt64 w = osghtonll( *((reinterpret_cast<const UInt64 *>(&value)) + 1) );

#if BYTE_ORDER == LITTLE_ENDIAN
    putToChannel(channel, &w, sizeof(UInt64));
    putToChannel(channel, &v, sizeof(UInt64));
#else
    putToChannel(channel, &v, sizeof(UInt64));
    putToChannel(channel, &w, sizeof(UInt64));
#endif
}

void GroupSockConnection::putValueToChannel(Channel channel, const std::string &value)
{
    UInt32 len = stringlen(value.c_str()) + 1;

    putValueToChannel(channel, len);

    if(len != 0)
    {
        putToChannel(channel, value.c_str(), len);
    }
}


typedef struct
{
    std::vector<UInt8> buffer;
    GroupSockConnection *the;
} ThreadData;

/** Write buffer
 *
 * Write blocksize and data.
 *
 **/
void GroupSockConnection::writeBuffer(void)
{
    //Real64 t = OSG::getSystemTime();

    UInt32 size = writeBufBegin()->getDataSize();

    if(size)
    {
        // write size to header
        reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->magic = 0x12345678;
        reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->id = ++_id;
        reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->size = osghtonl(size);
        reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->compressed_size = osghtonl(0);
        UInt8 *data = &_socketWriteBuffer[0];

        _write_buffer_bytes += size;

        Real64 t = OSG::getSystemTime();
        if(_compression == Connection::CCOMPRESSION_BEST && size > 1)
        {
            static std::vector<UInt8> compressed;
            compressed.resize(mz_compressBound(size) + sizeof(SocketBufferHeader));
            mz_ulong compressed_size = compressed.size() - sizeof(SocketBufferHeader);
            int r = mz_compress2((unsigned char *) &compressed[sizeof(SocketBufferHeader)], &compressed_size, (const unsigned char *) &_socketWriteBuffer[sizeof(SocketBufferHeader)], size, 1);
            if(r == 0 && compressed_size != 0 && compressed_size < size)
            {
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->magic = 0x12345678;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->id = reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->id;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->size = osghtonl(size);
                size = compressed_size;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->compressed_size = osghtonl(size);
                data = &compressed[0];
            }
            else
            {
                //printf("GroupSockConnection::writeBuffer: send uncompressed %d!\n", size);
            }
        }
        else if(_compression == Connection::CCOMPRESSION_FAST && size > 1)
        {
            static std::vector<UInt8> compressed;
            // ok we allocate size - 1, lzf_compress returns 0 when the compression failed,
            // in this case the compressed data is greater than the input data.
            //compressed.resize(size - 1 + sizeof(SocketBufferHeader));
            compressed.resize(LZ4_compressBound(size) + sizeof(SocketBufferHeader));
            int compressed_size = LZ4_compress((const char *) &_socketWriteBuffer[sizeof(SocketBufferHeader)], (char *) &compressed[sizeof(SocketBufferHeader)], size);

            if(compressed_size != 0 && compressed_size < size)
            {
                //printf("PointSockConnection::writeBuffer: compressed %d ==> %d\n", size, compressed_size);
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->magic = 0x12345678;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->id = reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->id;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->size = osghtonl(size);
                size = compressed_size;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->compressed_size = osghtonl(size);
                data = &compressed[0];
            }
            else
            {
                //printf("GroupSockConnection::writeBuffer: send uncompressed %d!\n", size);
            }
        }
        else
        {
            //printf("GroupSockConnection::writeBuffer: send uncompressed %d!\n", size);
        }

#ifdef MULTI_THREADED
        _tp.wait();

        static ThreadData *td = new ThreadData;
        td->buffer.resize(size + sizeof(SocketBufferHeader));
        memcpy(&td->buffer[0], data, size + sizeof(SocketBufferHeader));
        td->the = this;
        _tp.run(writeThread, static_cast<void *>(td));
#else
        GroupSockConnection::write(data, size + sizeof(SocketBufferHeader));
#endif

        _write_buffer_cbytes += size;
        _write_buffer_time += (OSG::getSystemTime() - t);
    }

    //printf("GroupSockConnection::writeBuffer: %f\n", OSG::getSystemTime() - t);
}

/*! write thread
 */
void GroupSockConnection::writeThread(void *arg)
{
    ThreadData *td = (ThreadData *) arg;
    td->the->GroupSockConnection::write(&td->buffer[0], td->buffer.size());
    // we use one thread only so one td instance is enough!
    //delete td;
}

/*-------------------------------------------------------------------------*/
/*                              static type                                */

ConnectionType GroupSockConnection::_type(
    &GroupSockConnection::create,
    "StreamSock");

/*-------------------------------------------------------------------------*/
/*                              cvs id's                                   */

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp       [] = "@(#)$Id: $";
    static Char8 cvsid_hpp       [] = OSG_GROUPSOCKCONNECTION_HEADER_CVSID;
}

