/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#ifndef _OSGGEOPUMPFACTORY_H_
#define _OSGGEOPUMPFACTORY_H_
#ifdef __sgi
#pragma once
#endif

//#define OSG_NV_BINDLESS_SUPPORT

#include <utility>
#include <OSGBaseTypes.h>
#include <OSGBaseFunctions.h>
#include <OSGGeometry.h>

OSG_BEGIN_NAMESPACE


typedef UInt64 GLuint64;

class Window;

#ifdef OSG_SUPPORT_NO_GEO_INTERFACE
class GeoPositionsInterface;
class GeoNormalsInterface;
class GeoColorsInterface;
class GeoTexCoordsInterface;
class GeoPTypesInterface;
class GeoPLengthsInterface;
class GeoIndicesInterface;
#endif

/*! \brief Selects a pump function for drawing the given geometry.
*/

class OSG_SYSTEMLIB_DLLMAPPING GeoPumpFactory
{
    /*==========================  PUBLIC  =================================*/
  public:

    typedef UInt32 Index;

    /*---------------------------------------------------------------------*/
    /*! \name                    Pump Types                                */
    /*! \{                                                                 */

    typedef void (*GeoPump)( Window *win, Geometry * geo );

    typedef void (*PartialGeoPump)( Window *win, Geometry * geo,
        UInt32 primtype, UInt32 firstvert, UInt32 nvert );

    typedef void (*InterfacePump)( Window *win,
        GeoPositionsInterface *pos, GeoNormalsInterface *norm,
        GeoColorsInterface *col, GeoColorsInterface *seccol,
        GeoTexCoordsInterface *texcoords,
        GeoTexCoordsInterface *texcoords1,
        GeoTexCoordsInterface *texcoords2,
        GeoTexCoordsInterface *texcoords3,
        GeoPTypesInterface *type, GeoPLengthsInterface *len,
        GeoIndicesInterface *ind, UInt16 *map, UInt16 nmap );

    typedef void (*PartialInterfacePump)( Window *win,
        GeoPositionsInterface *pos, GeoNormalsInterface *norm,
        GeoColorsInterface *col, GeoColorsInterface *seccol,
        GeoTexCoordsInterface *texcoords,
        GeoTexCoordsInterface *texcoords1,
        GeoTexCoordsInterface *texcoords2,
        GeoTexCoordsInterface *texcoords3,
        GeoPTypesInterface *type, GeoPLengthsInterface *len,
        GeoIndicesInterface *ind, UInt16 *map, UInt16 nmap,
        UInt32 primtype, UInt32 firstvert, UInt32 nvert );

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                    Class Get                                 */
    /*! \{                                                                 */

    static const char *getClassname(void) { return "GeoPumpFactory"; }

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   Destructors                                */
    /*! \{                                                                 */

    virtual ~GeoPumpFactory(void);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                       Get                                    */
    /*! \{                                                                 */

    Index                getIndex                (Geometry * geo          );

    GeoPump              getGeoPump              (Window *win, Index index);

    PartialGeoPump       getPartialGeoPump       (Window *win, Index index);

    InterfacePump        getInterfacePump        (Window *win, Index index);

    PartialInterfacePump getPartialInterfacePump (Window *win, Index index);


    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                       Access                                 */
    /*! \{                                                                 */

    static GeoPumpFactory *the( void );

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name               OpenGL Extension handling                      */
    /*! \{                                                                 */

    static UInt32           _extSecondaryColor;
    static UInt32           _extMultitexture;
    static UInt32           _extCompiledVertexArray;
    static UInt32           _extDrawRangeElements;
    static UInt32           _arbVBO;
    static UInt32           _arbVAO;


    static UInt32           _funcglSecondaryColorPointer;
    static UInt32           _funcglClientActiveTextureARB;
    static UInt32           _funcglLockArraysEXT;
    static UInt32           _funcglUnlockArraysEXT;
    static UInt32           _funcglDrawRangeElementsEXT;

    static UInt32           _funcglGenBuffersARB;
    static UInt32           _funcglDeleteBuffersARB;
    static UInt32           _funcglBindBufferARB;
    static UInt32           _funcglBufferDataARB;
    static UInt32           _funcglBufferSubDataARB;

    static UInt32           _funcglGenBuffers;
    static UInt32           _funcglDeleteBuffers;
    static UInt32           _funcglBindBuffer;
    static UInt32           _funcglBufferData;
    static UInt32           _funcglBufferSubData;

    static UInt32           _funcglGenVertexArrays;
    static UInt32           _funcglBindVertexArray;
    static UInt32           _funcglDeleteVertexArrays;

#ifdef OSG_NV_BINDLESS_SUPPORT
    // bindless VBO stuff
    static UInt32           _nvVertexBufferUnifiedMemory;
    static UInt32           _nvShaderBufferLoad;

    static UInt32           _funcglBufferAddressRangeNV;
    static UInt32           _funcglVertexFormatNV;
    static UInt32           _funcglNormalFormatNV;
    static UInt32           _funcglColorFormatNV;
    static UInt32           _funcglSecondaryColorFormatNV;
    static UInt32           _funcglTexCoordFormatNV;
    static UInt32           _funcglIndexFormatNV;

    static UInt32           _funcglMakeBufferResidentNV;
    static UInt32           _funcglGetBufferParameterui64vNV;
#endif

    /*! \}                                                                 */
    /*=========================  PROTECTED  ===============================*/
  protected:

    virtual Index numIndices(void);

    /*---------------------------------------------------------------------*/
    /*! \name                   Constructors                               */
    /*! \{                                                                 */

    GeoPumpFactory(void);

    /*! \}                                                                 */
    /*==========================  PRIVATE  ================================*/
  private:

    static char cvsid[];

    static GeoPumpFactory *_the;
    static GeoPump GeoPumps[131];

    /*---------------------------------------------------------------------*/
    /*! \name               OpenGL Extension handling                      */
    /*! \{                                                                 */

    static bool             glextInitFunction(void);
    static InitFuncWrapper _glextInitFuncWrapper;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                       The Pumps                              */
    /*! \{                                                                 */

    static void masterGeoPump(Window *win, Geometry *geo);

    static void masterPartialGeoPump(Window *win, Geometry *geo,
        UInt32 primtype, UInt32 firstvert, UInt32 nvert);

    static void masterInterfacePump(Window *win,
        GeoPositionsInterface *pos, GeoNormalsInterface *norm,
        GeoColorsInterface *col, GeoColorsInterface *seccol,
        GeoTexCoordsInterface *texcoords,
        GeoTexCoordsInterface *texcoords2,
        GeoTexCoordsInterface *texcoords3,
        GeoTexCoordsInterface *texcoords4,
        GeoPTypesInterface *type, GeoPLengthsInterface*len,
        GeoIndicesInterface *ind, UInt16 *map, UInt16 nmap);

    static void masterPartialInterfacePump(Window *win,
        GeoPositionsInterface *pos, GeoNormalsInterface *norm,
        GeoColorsInterface *col, GeoColorsInterface *seccol,
        GeoTexCoordsInterface *texcoords,
        GeoTexCoordsInterface *texcoords2,
        GeoTexCoordsInterface *texcoords3,
        GeoTexCoordsInterface *texcoords4,
        GeoPTypesInterface *type, GeoPLengthsInterface*len,
        GeoIndicesInterface *ind, UInt16 *map, UInt16 nmap,
        UInt32 primtype, UInt32 firstvert, UInt32 nvert);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   Constructors                               */
    /*! \{                                                                 */

    GeoPumpFactory(const GeoPumpFactory &source);

    /*! \}                                                                 */

    void operator =(const GeoPumpFactory &source);

};

class OSG_SYSTEMLIB_DLLMAPPING GeoVBO
{
public:

    GeoVBO(Window *win, Geometry *geo);
    ~GeoVBO();

    void clear(void);
    void update(void);
    void draw(void);

    inline GLuint getPositions(void) const;
    inline GLuint getNormals(void) const;
    inline GLuint getColors(void) const;
    inline GLuint getSecColors(void) const;
    inline GLuint getTexCoords(void) const;
    inline GLuint getTexCoords1(void) const;
    inline GLuint getTexCoords2(void) const;
    inline GLuint getTexCoords3(void) const;
    inline GLuint getTexCoords4(void) const;
    inline GLuint getTexCoords5(void) const;
    inline GLuint getTexCoords6(void) const;
    inline GLuint getTexCoords7(void) const;
    inline GLuint getIndices(void) const;

    inline void setDrawPropertiesMask(UInt16 draw_properties_mask);
    inline UInt16 getDrawPropertiesMask(void) const;

    static void showASides( bool);
    static void showBSides( bool);
private:

    enum SideShowFlag
    {
        SHOW_A_SIDES = 1 << 1,
        SHOW_B_SIDES = 1 << 2
    };
    GLuint _vao;
    GLuint _positions;
    GLuint _normals;
    GLuint _colors;
    GLuint _secColors;
    GLuint _texCoords;
    GLuint _texCoords1;
    GLuint _texCoords2;
    GLuint _texCoords3;
    GLuint _texCoords4;
    GLuint _texCoords5;
    GLuint _texCoords6;
    GLuint _texCoords7;
    GLuint _indices;

#ifdef OSG_NV_BINDLESS_SUPPORT
    GLuint64 _positions_addr;
    GLuint64 _normals_addr;
    GLuint64 _colors_addr;
    GLuint64 _secColors_addr;
    GLuint64 _texCoords_addr;
    GLuint64 _texCoords1_addr;
    GLuint64 _texCoords2_addr;
    GLuint64 _texCoords3_addr;
    GLuint64 _texCoords4_addr;
    GLuint64 _texCoords5_addr;
    GLuint64 _texCoords6_addr;
    GLuint64 _texCoords7_addr;
    GLuint64 _indices_addr;
    GLuint _positions_size;
    GLuint _normals_size;
    GLuint _colors_size;
    GLuint _secColors_size;
    GLuint _texCoords_size;
    GLuint _texCoords1_size;
    GLuint _texCoords2_size;
    GLuint _texCoords3_size;
    GLuint _texCoords4_size;
    GLuint _texCoords5_size;
    GLuint _texCoords6_size;
    GLuint _texCoords7_size;
    GLuint _indices_size;
#endif

    UInt32 _maxIndex;
    UInt32 _numIndices; // the number of indices in a shell
    UInt16 _valid_draw_properties_mask; // the valid draw properties for this geo
    UInt16 _draw_properties_mask;

    static UInt32   _render_sides; // bitmask storing the global render sides settings. bit 1 = A sides, bit 2 = B sides
    Window   *_win;
    Geometry *_geo;
    bool      _cleared;
};

OSG_END_NAMESPACE

#include "OSGGeoPumpFactory.inl"

#define OSGGEOPUMPFACTORY_HEADER_CVSID "@(#)$Id: $"

#endif /* _OSGGEOPUMPFACTORY_H_ */
