/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *               Copyright (C) 2000-2002 by the OpenSG Forum                 *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>

#include <OSGConfig.h>
#include <OSGDrawAction.h>
#include <OSGRenderAction.h>
#include <OSGIntersectAction.h>
#include <OSGIntersectActor.h>

#include "OSGTransform3D.h"

OSG_BEGIN_NAMESPACE

/***************************************************************************\
 *                            Description                                  *
\***************************************************************************/

/*! \class osg::Transform3D
Holds values for translation, rotation, scale, shearing and pivots. Constructs the finals transformation matrix out of the values.
*/

/***************************************************************************\
 *                           Class variables                               *
\***************************************************************************/

const OSG::BitVector Transform3D::DummyFieldMask = FlagsFieldMask;

/***************************************************************************\
 *                           Class methods                                 *
\***************************************************************************/

void Transform3D::initMethod (void)
{
    DrawAction::registerEnterDefault(
                                     getClassType(),
                                     osgTypedMethodFunctor2BaseCPtrRef<
                                     Action::ResultE,
                                     Transform3DPtr  ,
                                     CNodePtr               ,
                                     Action                *>(&Transform3D::drawEnter));

    DrawAction::registerLeaveDefault(
                                     getClassType(),
                                     osgTypedMethodFunctor2BaseCPtrRef<
                                     Action::ResultE,
                                     Transform3DPtr  ,
                                     CNodePtr               ,
                                     Action                *>(&Transform3D::drawLeave));


    RenderAction::registerEnterDefault(
                                       getClassType(),
                                       osgTypedMethodFunctor2BaseCPtrRef<
                                       Action::ResultE,
                                       Transform3DPtr  ,
                                       CNodePtr               ,
                                       Action                *>(&Transform3D::renderEnter));

    RenderAction::registerLeaveDefault(
                                       getClassType(),
                                       osgTypedMethodFunctor2BaseCPtrRef<
                                       Action::ResultE,
                                       Transform3DPtr  ,
                                       CNodePtr               ,
                                       Action                *>(&Transform3D::renderLeave));

    IntersectAction::registerEnterDefault(
                                          getClassType(),
                                          osgTypedMethodFunctor2BaseCPtrRef<
                                          Action::ResultE,
                                          Transform3DPtr  ,
                                          CNodePtr               ,
                                          Action                *>(&Transform3D::intersectEnter));

    IntersectAction::registerLeaveDefault(
                                          getClassType(),
                                          osgTypedMethodFunctor2BaseCPtrRef<
                                          Action::ResultE,
                                          Transform3DPtr  ,
                                          CNodePtr               ,
                                          Action                *>(&Transform3D::intersectLeave));

    IntersectActor::regClassEnter(
                                  osgTypedMethodFunctor2BaseCPtr<
                                  NewActionTypes::ResultE,
                                  Transform3DPtr  ,
                                  NodeCorePtr            ,
                                  ActorBase::FunctorArgumentType &>(&Transform3D::intersectActorEnter),
                                  getClassType());

    IntersectActor::regClassLeave(
                                  osgTypedMethodFunctor2BaseCPtr<
                                  NewActionTypes::ResultE,
                                  Transform3DPtr  ,
                                  NodeCorePtr            ,
                                  ActorBase::FunctorArgumentType &>(&Transform3D::intersectActorLeave),
                                  getClassType());
}


/***************************************************************************\
 *                           Instance methods                              *
\***************************************************************************/

/*-------------------------------------------------------------------------*\
 -  private                                                                 -
\*-------------------------------------------------------------------------*/

/*----------------------- constructors & destructors ----------------------*/

Transform3D::Transform3D(void) :
    Inherited()
{
    _sfScale.getValue().setValues(1.f, 1.f, 1.f);
    _sfEulerRotationOrder.setValue("xyz");
}

Transform3D::Transform3D(const Transform3D &source) :
    Inherited(source)
{
}

Transform3D::~Transform3D(void)
{
}

/*----------------------------- class specific ----------------------------*/

void Transform3D::changed(BitVector whichField, UInt32 origin)
{
    //Transform3DPtr ptr(*this);
    // This works when load
#if 0
    if( whichField & RotationFieldMask )
    {
        updateEulerByQuat();
    }
    if( whichField & RotationOrientationFieldMask )
    {
        updateEulerByQuatOrient();
    }
#else
    if( whichField & EulerRotationFieldMask )
    {
        //updateQuatByEuler();
    }
    if( whichField & EulerRotationOrientationFieldMask )
    {
        //updateQuatOrientByEuler();
    }
    if( whichField & EulerRotationOrderFieldMask )
    {
        //updateQuatOrientByEuler();
        //updateQuatByEuler();
    }
#endif

    if( (whichField == FieldBits::AllFields) || !(whichField & Inherited::MatrixFieldMask) )
    {
        Matrix final = computeFinalMatrix();

#if 0
        // already done in Transform::changed()!
        if( memcmp(final.getValues(), _identity.getValues(), sizeof(Matrix)) == 0 )
            _sfIsIdentity.setValue(true);
        else
            _sfIsIdentity.setValue(false);
#endif

        _sfMatrix.setValue(final);

#if 0
        // already done in Transform::changed()!
        invalidateVolume();
#endif
    }

    Inherited::changed(whichField, origin);
}

void Transform3D::addFlag(UInt64 flag)
{
    UInt64 flags = _sfFlags.getValue();
    if( !( flags & flag) )
        _sfFlags.setValue( flags + flag );
}

void Transform3D::subFlag(UInt64 flag)
{
    UInt64 flags = _sfFlags.getValue();
    if( flags & flag )
        _sfFlags.setValue( flags - flag );
}

void Transform3D::setDummy(bool value)
{
    if( value )
        addFlag(DUMMY);
    else
        subFlag(DUMMY);
}

bool Transform3D::getDummy(void)
{
    return (_sfFlags.getValue() & DUMMY);
}

Matrix Transform3D::computeFinalMatrix()
{
    Matrix final;
    final.setIdentity();

    Matrix r, ro, s, t, sh, sp, spInv, spt, rp, rpInf, rpt;
    t.setTranslate(getTranslation());
    s.setScale(getScale());
    getRotation().getValue(r);
    getRotationOrientation().getValue(ro);
    rp.setTranslate(getRotatePivot());
    rp.inverse(rpInf);
    rpt.setTranslate(getRotatePivotTranslation());
    sp.setTranslate(getScalePivot());
    sp.inverse(spInv);
    spt.setTranslate(getScalePivotTranslation());
    Vec3f shear = getShear();
    sh.setValue(1.0, 0.0, 0.0, 0.0,
                shear.x(), 1.0, 0.0, 0.0,
                shear.y(), shear.z(), 1.0, 0.0,
                0.0, 0.0, 0.0, 1.0);
    sh.transpose();

    final.mult(t);
    final.mult(rpt);
    final.mult(rp);
    final.mult(r);
    //final.mult(ro);
    final.mult(rpInf);
    final.mult(spt);
    final.mult(sp);
    final.mult(sh);
    final.mult(s);
    final.mult(spInv);

    return final;
}

Matrix Transform3D::getPivotMatrix() const
{
    Matrix r, ro, t, rp, rpInf, rpt;
    t.setTranslate(getTranslation());
    getRotation().getValue(r);
    getRotationOrientation().getValue(ro);
    rp.setTranslate(getRotatePivot());
    rp.inverse(rpInf);
    rpt.setTranslate(getRotatePivotTranslation());

    Matrix m = t;
    m.mult(rpt);
    m.mult(rp);
    m.mult(r);
    m.mult(ro);
    m.mult(rpInf);

    return m;
}

void Transform3D::reset()
{
    _sfTranslation.setValue(Vec3f());
    _sfRotation.setValue(Quaternion());
    _sfScale.setValue(Vec3f(1.0, 1.0, 1.0));
    _sfRotatePivot.setValue(Pnt3f());
    _sfScalePivot.setValue(Pnt3f());
    _sfRotationOrientation.setValue(Quaternion());
    _sfShear.setValue(Vec3f());
    _sfRotatePivotTranslation.setValue(Vec3f());
    _sfScalePivotTranslation.setValue(Vec3f());
    _sfCenter.setValue(Pnt3f());
    _sfEulerRotation.setValue(Vec3f());
    _sfEulerRotationOrientation.setValue(Vec3f());
    _sfEulerRotationOrder.setValue("xyz");

    // ensure the matrix is identity
    Matrix m;
    m.setIdentity();
    _sfMatrix.setValue(m);
    _sfIsIdentity.setValue(true);
}

void Transform3D::updateQuatByEuler(void)
{
    Quaternion q = convertEulerToQuat(_sfEulerRotation.getValue(), true);
    _sfRotation.setValue(q);
}

void Transform3D::updateQuatOrientByEuler(void)
{
    Quaternion q = convertEulerToQuat(_sfEulerRotationOrientation.getValue(), false);
    _sfRotationOrientation.setValue(q);
}

Quaternion Transform3D::convertEulerToQuat(const Vec3f& euler, bool addOrientation)
{
    Vec3f xAxis(1.0, 0.0, 0.0);
    Vec3f yAxis(0.0, 1.0, 0.0);
    Vec3f zAxis(0.0, 0.0, 1.0);

    if( addOrientation )
    {
        Quaternion rOrient = _sfRotationOrientation.getValue();

        // set up the orientation
        rOrient.multVec(xAxis, xAxis);
        rOrient.multVec(yAxis, yAxis);
        rOrient.multVec(zAxis, zAxis);
    }

    // rotation for the single axises
    Quaternion qLocal, qGlobal, qx, qy, qz;
    qx.setValueAsAxisDeg(xAxis, euler.x());
    qy.setValueAsAxisDeg(yAxis, euler.y());
    qz.setValueAsAxisDeg(zAxis, euler.z());

    Matrix rxMat, ryMat, rzMat;
    qx.getValue(rxMat);
    qy.getValue(ryMat);
    qz.getValue(rzMat);

    Matrix rotMat;

    std::string order = _sfEulerRotationOrder.getValue();
    if( order == "xyz" )
    {
        rotMat = rzMat;
        rotMat.mult(ryMat);
        rotMat.mult(rxMat);

        qx.mult(qy);
        qx.mult(qz);
        qLocal = qx;
    }
    else if( order == "yzx" )
    {
        rotMat = rxMat;
        rotMat.mult(rzMat);
        rotMat.mult(ryMat);

        qy.mult(qz);
        qy.mult(qx);
        qLocal = qy;
    }
    else if( order == "zxy" )
    {
        rotMat = ryMat;
        rotMat.mult(rxMat);
        rotMat.mult(rzMat);

        qz.mult(qx);
        qz.mult(qy);
        qLocal = qz;
    }
    else if( order == "xzy" )
    {
        rotMat = ryMat;
        rotMat.mult(rzMat);
        rotMat.mult(rxMat);

        qx.mult(qz);
        qx.mult(qy);
        qLocal = qx;
    }
    else if( order == "yxz" )
    {
        rotMat = rzMat;
        rotMat.mult(rxMat);
        rotMat.mult(ryMat);

        qy.mult(qx);
        qy.mult(qz);
        qLocal = qy;
    }
    else if( order == "zyx" )
    {
        rotMat = rxMat;
        rotMat.mult(ryMat);
        rotMat.mult(rzMat);

        qz.mult(qy);
        qz.mult(qx);
        qLocal = qz;
    }

    qGlobal.setValue(rotMat);

    if( _sfRotationMode.getValue() == 0 )
        return qGlobal;
    else
        return qLocal;
}

void Transform3D::updateEulerByQuat(void)
{
    Vec3f euler = convertQuatToEuler(_sfRotation.getValue(), true);
    _sfEulerRotation.setValue(euler);
}

void Transform3D::updateEulerByQuatOrient(void)
{
    Vec3f euler = convertQuatToEuler(_sfRotationOrientation.getValue(), false);
    _sfEulerRotationOrientation.setValue(euler);
}

Vec3f Transform3D::convertQuatToEuler(const Quaternion& value, bool addOrientation)
{
    Real32 rx = 0.f;
    Real32 ry = 0.f;
    Real32 rz = 0.f;

    Quaternion q;

    if( addOrientation )
    {
        Quaternion ro = _sfRotationOrientation.getValue();
        Quaternion tr = value;
        q = ro.inverse();
        q.mult(tr);
        q.mult(ro);
    }
    else
    {
        q = value;
    }

    if( _sfRotationMode.getValue() == 0 )
    {
        Real32 sqw = q.w()*q.w();
        Real32 sqx = q.x()*q.x();
        Real32 sqy = q.y()*q.y();
        Real32 sqz = q.z()*q.z();

        rx = rad2deg(atan2f(2.0 * (q.y()*q.z() + q.x()*q.w()),(-sqx - sqy + sqz + sqw)));
        ry = rad2deg(asinf( std::max(-1.0f, std::min( 1.0f, -2.0f * (q.x()*q.z() - q.y()*q.w())))));
        rz = rad2deg(atan2f(2.0 * (q.x()*q.y() + q.z()*q.w()),(sqx - sqy - sqz + sqw)));
    }
    else
    {
        Real32 p0 = q.w();
        Real32 p1 = q.x();
        Real32 p2 = q.y();
        Real32 p3 = q.z();

        Real32 nom0 = 2.0 * ( p0 * p1 - p2 * p3);
        Real32 denom0 = 1.0 - 2.0 * ( p1 * p1 + p2 * p2);
        Real32 nom1 = 2.0 * ( p0 * p2 + p1 * p3);
        Real32 nom2 =  2.0 * ( p0 * p3 - p1 * p2);
        Real32 denom2 = 1.0 - 2.0 * ( p2 * p2 + p3 * p3);

        rx = rad2deg( atan2f( nom0, denom0));
        ry = rad2deg( asinf(std::max(-1.0f, std::min( 1.0f, nom1))));
        rz = rad2deg( atan2f(nom2,denom2));
    }

    return Vec3f(rx, ry, rz);
}

void Transform3D::dump(      UInt32    ,
                         const BitVector ) const
{
    SLOG << "Dump Transform3D NI" << std::endl;
}


/*------------------------------------------------------------------------*/
/*                              cvs id's                                  */

#ifdef OSG_SGI_CC
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp       [] = "@(#)$Id: FCTemplate_cpp.h,v 1.20 2006/03/16 17:01:53 dirk Exp $";
    static Char8 cvsid_hpp       [] = OSGTRANSFORM3DBASE_HEADER_CVSID;
    static Char8 cvsid_inl       [] = OSGTRANSFORM3DBASE_INLINE_CVSID;

    static Char8 cvsid_fields_hpp[] = OSGTRANSFORM3DFIELDS_HEADER_CVSID;
}

#ifdef __sgi
#pragma reset woff 1174
#endif

OSG_END_NAMESPACE

